package ru.dbotthepony.mc.otm.core.math

import it.unimi.dsi.fastutil.objects.ObjectArrayList
import ru.dbotthepony.mc.otm.core.random
import java.util.random.RandomGenerator
import kotlin.math.min

interface Cluster<V : Any> {
	val values: List<V>
	val center: V

	data class Impl<V : Any>(override val values: List<V>, override val center: V) : Cluster<V>
}

private class ClusterValue<V : Comparable<V>>(val value: V, var cluster: MutableCluster<V>, var error: V) {
	inline fun updateError(abs: (V) -> V, minus: (V, V) -> V) {
		error = abs(minus(cluster.center, value))
	}

	inline fun maybeSwitchCluster(cluster: MutableCluster<V>, abs: (V) -> V, minus: (V, V) -> V): Boolean {
		if (cluster == this.cluster) return false

		val newError = abs(minus(cluster.center, value))

		if (newError < error) {
			error = newError
			this.cluster.values.remove(this)
			cluster.values.add(this)
			this.cluster = cluster
			return true
		} else {
			return false
		}
	}
}

private class MutableCluster<V : Comparable<V>>(var center: V, expectedSize: Int) {
	val values = ObjectArrayList<ClusterValue<V>>(min(100, expectedSize))

	inline fun calculateCenter(identity: V, plus: (V, V) -> V, minus: (V, V) -> V, divInt: (V, Int) -> V, abs: (V) -> V): Boolean {
		if (values.isEmpty()) return false

		var value = identity
		values.forEach { value = plus(value, it.value) }
		val old = center
		center = divInt(value, values.size)

		if (old != center) {
			values.forEach { it.updateError(abs, minus) }
		}

		return true
	}
}

private inline fun <V : Comparable<V>> Iterable<V>.clusterize(
	random: RandomGenerator,
	initialClusters: Int = 1,
	identity: V,
	plus: (V, V) -> V,
	minus: (V, V) -> V,
	divInt: (V, Int) -> V,
	abs: (V) -> V,
	heuristics: (min: V, max: V, error: V) -> Boolean,
): List<Cluster<V>> {
	val itr = iterator()

	if (!itr.hasNext())
		return listOf()

	var expectedSize = 1
	var min = itr.next()
	var max = min

	while (itr.hasNext()) {
		val value = itr.next()

		min = minOf(min, value)
		max = maxOf(max, value)
		expectedSize++
	}

	if (min == max) {
		return listOf(Cluster.Impl(listOf(min), min))
	}

	var targetClusters = initialClusters
	val clusters = ObjectArrayList<MutableCluster<V>>(initialClusters)
	val values = ObjectArrayList<ClusterValue<V>>(expectedSize)

	while (true) {
		clusters.clear()
		values.clear()

		var converged = false
		var oversaturation = false

		for (i in 0 until targetClusters) {
			clusters.add(MutableCluster(identity, expectedSize))
		}

		for (value in this) {
			val cluster = clusters.random(random)
			val wrapped = ClusterValue(value, cluster, value)
			values.add(wrapped)
			cluster.values.add(wrapped)
		}

		clusters.forEach { it.calculateCenter(identity, plus, minus, divInt, abs) }

		while (!converged) {
			converged = true

			for (value in values) {
				clusters.forEach { converged = !value.maybeSwitchCluster(it, abs, minus) && converged }
			}

			if (!converged) {
				val emptyClusters = ObjectArrayList<MutableCluster<V>>()
				val citr = clusters.iterator()

				for (cluster in citr) {
					if (!cluster.calculateCenter(identity, plus, minus, divInt, abs)) {
						emptyClusters.add(cluster)
						citr.remove()
					}
				}

				for (cluster in emptyClusters) {
					var candidate: ClusterValue<V>? = null

					for (eCluser in clusters) {
						if (eCluser.values.size > 1) {
							for (value in eCluser.values) {
								if (candidate == null || candidate.error < value.error) {
									candidate = value
								}
							}
						}
					}

					if (candidate != null) {
						cluster.values.add(candidate)
						candidate.cluster.values.remove(candidate)
						candidate.cluster = cluster
						cluster.center = candidate.value
						candidate.error = identity
						clusters.add(cluster)
					} else {
						oversaturation = true
						break
					}
				}
			}
		}

		val maxError = values.maxOf { it.error }

		if (!oversaturation && targetClusters < values.size && heuristics(min, max, maxError)) {
			targetClusters++
		} else {
			return clusters.stream()
				.filter { it.values.isNotEmpty() }
				.map { Cluster.Impl(it.values.map { it.value }, it.center) }
				.sorted { o1, o2 -> o2.values.size.compareTo(o1.values.size) } // большие кластеры должны идти первыми
				.toList()
		}
	}
}

// TODO: could use some tweaking
private val DECIMAL_ERROR_TOLERANCE = Decimal("0.02")

fun Iterable<Decimal>.clusterize(random: RandomGenerator, clusters: Int? = null): List<Cluster<Decimal>> {
	return clusterize(random, clusters ?: 1, Decimal.ZERO, Decimal::plus, Decimal::minus, Decimal::div, Decimal::absoluteValue) { min, max, error ->
		if (clusters != null)
			false
		else
			error / (max - min) >= DECIMAL_ERROR_TOLERANCE
	}
}
